from django.shortcuts import redirect
from django.utils.deprecation import MiddlewareMixin


class MFAEnforceMiddleware(MiddlewareMixin):
    def process_view(self, request, view_func, view_args, view_kwargs):
        if (
            not getattr(view_func, 'mfa_public', False)
            and request.user.is_authenticated
            and not request.user.mfakey_set.exists()
        ):
            return redirect('mfa:list')



from django.conf import settings
from django.db import models


class MFAKey(models.Model):
    id = models.BigAutoField(primary_key=True, serialize=False, verbose_name='ID')
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE
    )
    method = models.CharField(max_length=8, choices=[
        ('FIDO2', 'FIDO2'),
        ('TOTP', 'TOTP'),
        ('recovery', 'recovery'),
    ])
    name = models.CharField(max_length=32)
    secret = models.TextField()
    # replay protection
    last_code = models.CharField(max_length=32, blank=True)



from django import forms
from django.utils.translation import gettext_lazy as _


class MFABaseForm(forms.Form):
    code = forms.CharField(label=_('Authentication code'))

    def __init__(self, validate_code=None, **kwargs):
        self.validate_code = validate_code
        super().__init__(**kwargs)
        self.fields['code'].widget.attrs.update({'autocomplete': 'one-time-code'})

    def clean(self):
        cleaned_data = super().clean()
        code = cleaned_data.get('code')
        if code:
            try:
                cleaned_data['secret'] = self.validate_code(code)
            except ValueError as e:
                raise forms.ValidationError(_('Validation failed')) from e
        return cleaned_data


class MFAAuthForm(MFABaseForm):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fields['code'].widget.attrs.update({'autofocus': True})


class MFACreateForm(MFABaseForm):
    name = forms.CharField(label=_('Name for this key'), max_length=32)



from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth import login
from django.contrib.auth import user_login_failed
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.views.generic import DeleteView
from django.views.generic import ListView

from . import settings
from .decorators import login_not_required
from .decorators import stronghold_login_not_required
from .forms import MFAAuthForm
from .forms import MFACreateForm
from .mail import send_mail
from .mixins import MFAFormView
from .models import MFAKey


class LoginView(DjangoLoginView):
    def no_key_exists(self, form):
        return super().form_valid(form)

    def form_valid(self, form):
        user = form.get_user()
        if not user.mfakey_set.exists():
            return self.no_key_exists(form)

        self.request.session['mfa_user'] = {
            'pk': user.pk,
            'backend': user.backend,
        }
        self.request.session['mfa_success_url'] = self.get_success_url()
        for method in settings.METHODS:
            if user.mfakey_set.filter(method=method).exists():
                return redirect('mfa:auth', method)


class MFAListView(LoginRequiredMixin, ListView):
    model = MFAKey

    def get_queryset(self):
        return super().get_queryset().filter(user=self.request.user)


class MFADeleteView(LoginRequiredMixin, DeleteView):
    model = MFAKey

    def get_queryset(self):
        return super().get_queryset().filter(user=self.request.user)

    def get_success_url(self):
        return reverse('mfa:list')


class MFACreateView(LoginRequiredMixin, MFAFormView):
    form_class = MFACreateForm

    def get_template_names(self):
        return f'mfa/create_{self.method.name}.html'

    def get_success_url(self):
        return reverse('mfa:list')

    def begin(self):
        return self.method.register_begin(self.request.user)

    def complete(self, code):
        return self.method.register_complete(self.challenge[1], code)

    def form_valid(self, form):
        MFAKey.objects.create(
            user=self.request.user,
            method=self.method.name,
            name=form.cleaned_data['name'],
            secret=form.cleaned_data['secret'],
        )
        messages.success(self.request, _('Key was created successfully!'))
        return super().form_valid(form)


@method_decorator(login_not_required, name='dispatch')
@method_decorator(stronghold_login_not_required, name='dispatch')
class MFAAuthView(MFAFormView):
    form_class = MFAAuthForm

    def get_template_names(self):
        return f'mfa/auth_{self.method.name}.html'

    def get_success_url(self):
        success_url = self.request.session.pop('mfa_success_url')
        if self.method.name == 'recovery':
            return reverse('mfa:list')
        else:
            return success_url

    @cached_property
    def user(self):
        try:
            user_data = self.request.session['mfa_user']
        except KeyError as e:
            raise Http404 from e
        User = get_user_model()
        user = get_object_or_404(User, pk=user_data['pk'])
        user.backend = user_data['backend']
        return user

    def begin(self):
        return self.method.authenticate_begin(self.user)

    def complete(self, code):
        return self.method.authenticate_complete(
            self.challenge[1], self.user, code,
        )

    def form_invalid(self, form):
        user_login_failed.send(
            sender=__name__,
            credentials={'username': self.user.get_username()},
            request=self.request,
        )
        send_mail(self.user, self.method)
        return super().form_invalid(form)

    def form_valid(self, form):
        login(self.request, self.user)
        del self.request.session['mfa_user']
        return super().form_valid(form)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mfa", "0003_alter_mfakey_method"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mfakey",
            name="id",
            field=models.BigAutoField(
                primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



# Generated by Django 3.2.4 on 2021-06-20 22:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mfa', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='mfakey',
            name='last_code',
            field=models.CharField(blank=True, max_length=32),
        ),
    ]



# Generated by Django 3.2.4 on 2021-07-31 06:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mfa', '0002_mfakey_last_code'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mfakey',
            name='method',
            field=models.CharField(choices=[('FIDO2', 'FIDO2'), ('TOTP', 'TOTP'), ('recovery', 'recovery')], max_length=8),
        ),
    ]



from fido2 import cbor
from fido2.features import webauthn_json_mapping
from fido2.server import Fido2Server
from fido2.utils import websafe_decode
from fido2.utils import websafe_encode
from fido2.webauthn import AttestationObject
from fido2.webauthn import AttestedCredentialData
from fido2.webauthn import AuthenticatorData
from fido2.webauthn import CollectedClientData
from fido2.webauthn import PublicKeyCredentialRpEntity

from .. import settings

name = 'FIDO2'

webauthn_json_mapping.enabled = False
fido2 = Fido2Server(
    PublicKeyCredentialRpEntity(id=settings.DOMAIN, name=settings.SITE_TITLE),
)


def encode(data):
    return cbor.encode(data).hex()


def decode(s):
    return cbor.decode(bytes.fromhex(s))


def get_credentials(user):
    keys = user.mfakey_set.filter(method=name)
    return [AttestedCredentialData(websafe_decode(key.secret)) for key in keys]


def register_begin(user):
    registration_data, state = fido2.register_begin(
        {
            'id': str(user.id).encode('utf-8'),
            'name': user.get_username(),
            'displayName': user.get_full_name(),
        },
        get_credentials(user),
        user_verification=settings.FIDO2_USER_VERIFICATION,
    )
    return encode(registration_data), state


def register_complete(state, request_data):
    data = decode(request_data)
    auth_data = fido2.register_complete(
        state,
        CollectedClientData(data['clientData']),
        AttestationObject(data['attestationObject']),
    )
    return websafe_encode(auth_data.credential_data)


def authenticate_begin(user):
    credentials = get_credentials(user)
    auth_data, state = fido2.authenticate_begin(
        credentials,
        user_verification=settings.FIDO2_USER_VERIFICATION,
    )
    return encode(auth_data), state


def authenticate_complete(state, user, request_data):
    data = decode(request_data)
    fido2.authenticate_complete(
        state,
        get_credentials(user),
        data['credentialId'],
        CollectedClientData(data['clientData']),
        AuthenticatorData(data['authenticatorData']),
        data['signature'],
    )



from django.http import Http404
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic import FormView

from . import settings
from .methods import fido2
from .methods import recovery
from .methods import totp


class MFAFormView(FormView):
    @property
    def method(self):
        if self.kwargs['method'] in settings.METHODS:
            if self.kwargs['method'] == 'FIDO2':
                return fido2
            elif self.kwargs['method'] == 'TOTP':
                return totp
            elif self.kwargs['method'] == 'recovery':
                return recovery
        raise Http404

    @property
    def challenge(self):
        try:
            return self.request.session['mfa_challenge']
        except KeyError as e:
            raise Http404 from e

    def begin(self):
        raise NotImplementedError  # pragma: no cover

    def complete(self, code):
        raise NotImplementedError  # pragma: no cover

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if 'mfa_data' not in context:
            data, state = self.begin()
            self.request.session['mfa_challenge'] = (data, state)
            context['mfa_data'] = data
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['validate_code'] = self.complete
        return kwargs

    def form_invalid(self, form):
        # do not generate a new challenge
        return self.render_to_response(self.get_context_data(
            form=form, mfa_data=self.challenge[0]
        ))

    def form_valid(self, form):
        del self.request.session['mfa_challenge']
        return super().form_valid(form)



from django.core.mail import EmailMultiAlternatives
from django.template import TemplateDoesNotExist
from django.template import loader

from . import settings

SUBJECT_TEMPLATE = 'mfa/login_failed_subject.txt'
BODY_TEMPLATE = 'mfa/login_failed_email.txt'
HTML_TEMPLATE = 'mfa/login_failed_email.html'


def send_mail(user, method):
    email_field_name = user.get_email_field_name()
    user_email = getattr(user, email_field_name)
    if not user_email:
        return 0

    context = {
        'email': user_email,
        'domain': settings.DOMAIN,
        'site_name': settings.SITE_TITLE,
        'user': user,
        'method': method.name,
    }

    try:
        subject = loader.render_to_string(SUBJECT_TEMPLATE, context)
        subject = ' '.join(subject.splitlines()).strip()
        body = loader.render_to_string(BODY_TEMPLATE, context)
    except TemplateDoesNotExist:
        return 0

    message = EmailMultiAlternatives(
        subject,
        body,
        to=[user_email],
        headers={'Auto-Submitted': 'auto-generated'},
    )

    try:
        html_body = loader.render_to_string(HTML_TEMPLATE, context)
        message.attach_alternative(html_body, 'text/html')
    except TemplateDoesNotExist:
        pass

    return message.send()



import pyotp
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import User
from django.core import mail
from django.test import TestCase
from fido2.server import _verify_origin_for_rp

from mfa.mail import send_mail
from mfa.methods import fido2
from mfa.models import MFAKey
from mfa.templatetags.mfa import get_qrcode


class MFATestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('test', password='password')

    def login(self):
        return self.client.post('/login/', {
            'username': 'test',
            'password': 'password',
        })

    def assert_not_logged_in(self):
        res = self.client.get('/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/login/?next=/')


class TOTPAuthViewTest(MFATestCase):
    def setUp(self):
        super().setUp()
        self.key = MFAKey.objects.create(
            user=self.user,
            method='TOTP',
            name='test',
            secret=pyotp.random_base32(),
        )
        self.totp = pyotp.TOTP(self.key.secret)

    def test_happy_flow(self):
        res = self.login()
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/auth/TOTP/')

        res = self.client.get('/mfa/auth/TOTP/')
        self.assertEqual(res.status_code, 200)

        res = self.client.post('/mfa/auth/TOTP/', {'code': self.totp.now()})
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/')

        res = self.client.get('/')
        self.assertEqual(res.status_code, 204)

    def test_invalid_method(self):
        self.login()
        res = self.client.get('/mfa/auth/INVALID/')
        self.assertEqual(res.status_code, 404)

    def test_not_logged_in_before_mfa_auth(self):
        self.login()
        self.assert_not_logged_in()

    def test_no_auth_without_login(self):
        res = self.client.get('/mfa/auth/TOTP/')
        self.assertEqual(res.status_code, 404)

    def test_no_auth_without_challenge(self):
        self.login()

        res = self.client.post('/mfa/auth/TOTP/', {'code': self.totp.now()})
        self.assertEqual(res.status_code, 404)
        self.assert_not_logged_in()

    def test_wrong_code(self):
        self.login()

        res = self.client.get('/mfa/auth/TOTP/')
        self.assertEqual(res.status_code, 200)

        res = self.client.post('/mfa/auth/TOTP/', {'code': 'invalid'})
        self.assertEqual(res.status_code, 200)
        self.assert_not_logged_in()


class TOTPCreateViewTest(MFATestCase):
    def test_happy_flow(self):
        self.client.force_login(self.user)

        res = self.client.get('/mfa/create/TOTP/')
        self.assertEqual(res.status_code, 200)
        totp = pyotp.TOTP(res.context['mfa_data']['secret'])

        res = self.client.post('/mfa/create/TOTP/', {
            'name': 'test',
            'code': totp.now()
        })
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/')

        self.assertEqual(MFAKey.objects.count(), 1)

    def test_not_logged_in(self):
        res = self.client.get('/mfa/create/TOTP/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/login/?next=/mfa/create/TOTP/')

    def test_no_create_without_challenge(self):
        self.client.force_login(self.user)

        res = self.client.post('/mfa/create/TOTP/', {
            'name': 'test',
            'code': 'invalid',
        })
        self.assertEqual(res.status_code, 404)
        self.assertEqual(MFAKey.objects.count(), 0)

    def test_wrong_code(self):
        self.client.force_login(self.user)

        res = self.client.get('/mfa/create/TOTP/')

        res = self.client.post('/mfa/create/TOTP/', {
            'name': 'test',
            'code': 'invalid',
        })
        self.assertEqual(res.status_code, 200)
        self.assertEqual(MFAKey.objects.count(), 0)

    def test_new_challenge_on_get(self):
        self.client.force_login(self.user)

        res = self.client.get('/mfa/create/TOTP/')
        secret1 = res.context['mfa_data']['secret']

        res = self.client.get('/mfa/create/TOTP/')
        secret2 = res.context['mfa_data']['secret']

        self.assertNotEqual(secret1, secret2)

        totp = pyotp.TOTP(secret1)
        self.client.post('/mfa/create/TOTP/', {
            'name': 'test',
            'code': totp.now()
        })
        self.assertEqual(MFAKey.objects.count(), 0)

    def test_keep_challenge_on_validation(self):
        self.client.force_login(self.user)

        res = self.client.get('/mfa/create/TOTP/')
        secret1 = res.context['mfa_data']['secret']

        res = self.client.post('/mfa/create/TOTP/', {
            'name': 'test',
            'code': 'invalid',
        })
        secret2 = res.context['mfa_data']['secret']
        self.assertEqual(secret1, secret2)

        totp = pyotp.TOTP(secret1)
        self.client.post('/mfa/create/TOTP/', {
            'name': 'test',
            'code': totp.now()
        })
        self.assertEqual(MFAKey.objects.count(), 1)


class FIDO2Test(MFATestCase):
    # I have no clue how to simulate a FIDO2 authenticator,
    # so these are just some smoke tests.

    def test_login_redirect(self):
        self.key = MFAKey.objects.create(
            user=self.user,
            method='FIDO2',
            name='test',
            secret='mock',
        )
        res = self.login()
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/auth/FIDO2/')

    def test_create(self):
        self.client.force_login(self.user)
        res = self.client.get('/mfa/create/FIDO2/')
        self.assertEqual(res.status_code, 200)

    def test_encode(self):
        self.assertEqual(fido2.encode({'foo': [1, 2]}), 'a163666f6f820102')

    def test_decode(self):
        self.assertEqual(fido2.decode('a163666f6f820102'), {'foo': [1, 2]})

    def test_origin_https(self):
        for domain, value, expected in [
            ('example.com', 'https://example.com', True),
            ('example.com', 'http://example.com', False),
            ('example.com', 'http://localhost:8000', False),
            ('localhost', 'https://example.com', False),
            ('localhost', 'http://localhost:8000', True),
            ('localhost', 'http://127.0.0.1', False),
            ('localhost', 'http://foo.localhost', True),
            ('127.0.0.1', 'http://127.0.0.1', False),
            ('foo.localhost', 'http://foo.localhost', True),
        ]:
            with self.subTest(domain=domain, value=value):
                verify = _verify_origin_for_rp(domain)
                self.assertEqual(verify(value), expected)


class RecoveryTest(MFATestCase):
    def test_create(self):
        self.client.force_login(self.user)

        res = self.client.get('/mfa/create/recovery/')
        self.assertEqual(res.status_code, 200)
        code = res.context['mfa_data']['code']
        self.assertEqual(len(code), 11)

        res = self.client.post('/mfa/create/recovery/', {
            'name': 'test',
            'code': 'invalid',
        })
        self.assertEqual(res.status_code, 200)

        res = self.client.post('/mfa/create/recovery/', {
            'name': 'test',
            'code': code,
        })
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/')

        self.assertEqual(MFAKey.objects.count(), 1)

    def test_authenticate(self):
        MFAKey.objects.create(
            user=self.user,
            method='FIDO2',
            name='test',
            secret='mock',
        )
        MFAKey.objects.create(
            user=self.user,
            method='recovery',
            name='recovery',
            secret=make_password('123456'),
        )

        res = self.login()
        self.assertEqual(res.status_code, 302)

        res = self.client.get('/mfa/auth/recovery/')
        self.assertEqual(res.status_code, 200)

        res = self.client.post('/mfa/auth/recovery/', {'code': 'invalid'})
        self.assertEqual(res.status_code, 200)

        res = self.client.post('/mfa/auth/recovery/', {'code': '123456'})
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/')

        res = self.client.get('/')
        self.assertEqual(res.status_code, 204)

        self.assertEqual(MFAKey.objects.count(), 1)


class MFAEnforceMiddlewareTest(MFATestCase):
    def test_redirect(self):
        self.login()
        res = self.client.get('/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/')

    def test_public(self):
        self.login()
        res = self.client.post('/logout/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/')


class PatchAdminTest(TestCase):
    def test_root(self):
        res = self.client.get('/admin/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/admin/login/?next=/admin/')

        res = self.client.get(res.url)
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/login/?next=/admin/')

    def test_app(self):
        res = self.client.get('/admin/mfa/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/admin/login/?next=/admin/mfa/')

        res = self.client.get(res.url)
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/login/?next=/admin/mfa/')

    def test_login(self):
        res = self.client.get('/admin/login/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/login/?next=/admin/')


class ListViewTest(MFATestCase):
    def test_list_view(self):
        self.client.force_login(self.user)
        MFAKey.objects.create(
            user=self.user,
            method='recovery',
            name='recovery',
            secret=make_password('123456'),
        )
        res = self.client.get('/mfa/')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.content.count(b'<li>'), 1)


class DeleteViewTest(MFATestCase):
    def test_delete_view(self):
        self.client.force_login(self.user)
        key = MFAKey.objects.create(
            user=self.user,
            method='recovery',
            name='recovery',
            secret=make_password('123456'),
        )
        res = self.client.post(f'/mfa/{key.pk}/delete/')
        self.assertEqual(res.status_code, 302)
        self.assertEqual(res.url, '/mfa/')
        self.assertEqual(MFAKey.objects.filter(pk=key.pk).count(), 0)


class QRCodeTest(TestCase):
    def test_is_svg(self):
        code = get_qrcode('some_data')
        self.assertTrue(code.startswith('<svg'))
        self.assertTrue(code.endswith('</svg>'))


class MailTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            'test', password='password', email='test@example.com'
        )

    def test_send_mail(self):
        count = send_mail(self.user, fido2)
        self.assertEqual(count, 1)

        message = mail.outbox[0]
        self.assertEqual(message.to, ['test@example.com'])
        self.assertEqual(message.subject, 'Attempted login to Tests using a wrong two-factor authentication code')  # noqa
        self.assertEqual(message.body, """Dear test,

We detected an attempt to log in to your account on Tests
(localhost) using a wrong two-factor authentication code. This means someone
managed to enter the correct password, but failed at FIDO2.

If this was you and you entered a wrong two-factor authentication code by
accident, you may ignore this email.

If this was not you, we strongly recommend to change your password.
""")



from django.urls import path

from .decorators import public
from .views import MFAAuthView
from .views import MFACreateView
from .views import MFADeleteView
from .views import MFAListView

app_name = 'mfa'
urlpatterns = [
    path('', public(MFAListView.as_view()), name='list'),
    path('<int:pk>/delete/', public(MFADeleteView.as_view()), name='delete'),
    path('create/<method>/', public(MFACreateView.as_view()), name='create'),
    path('auth/<method>/', MFAAuthView.as_view(), name='auth'),
]



import pyotp
from django.contrib.auth.hashers import check_password
from django.contrib.auth.hashers import make_password

name = 'recovery'


def register_begin(user):
    secret = pyotp.random_base32()
    totp = pyotp.TOTP(secret, digits=10)
    code = totp.now()
    code = f'{code[:5]}-{code[5:]}'
    state = make_password(code)
    return {'code': code}, state


def register_complete(state, request_data):
    if not check_password(request_data, state):
        raise ValueError
    return state


def authenticate_begin(user):
    return None, None


def authenticate_complete(state, user, request_data):
    keys = user.mfakey_set.filter(method=name)
    for key in keys:
        if check_password(request_data, key.secret):
            key.delete()
            return
    raise ValueError



from django.conf import settings

# The actual domain (e.g. "example.com") that this is hosted on
# FIDO2 will fail if this is not set correctly
DOMAIN = settings.MFA_DOMAIN

# Site title that is shown in authenticator apps
SITE_TITLE = settings.MFA_SITE_TITLE

# Available authentication methods in order of relevance
METHODS = getattr(settings, 'MFA_METHODS', ['FIDO2', 'TOTP', 'recovery'])

# `valid_window` parameter passed to PyOTP's verify method
# See https://pyauth.github.io/pyotp/#pyotp.totp.TOTP.verify
TOTP_VALID_WINDOW = getattr(settings, 'MFA_TOTP_VALID_WINDOW', 0)

# `user_verification` parameter passed to python-fido2
# See https://www.w3.org/TR/webauthn/#enum-userVerificationRequirement
FIDO2_USER_VERIFICATION = getattr(settings, 'MFA_FIDO2_USER_VERIFICATION', None)



from django.apps import AppConfig


class TwoFactorConfig(AppConfig):
    name = 'mfa'
    verbose_name = 'Multi Factor Authentication'

    def ready(self):
        from .admin import patch_admin
        patch_admin()



import pyotp

from .. import settings

name = 'TOTP'


def register_begin(user):
    secret = pyotp.random_base32()
    totp = pyotp.TOTP(secret)
    url = totp.provisioning_uri(
        user.get_username(),
        issuer_name=settings.SITE_TITLE,
    )
    return {'url': url, 'secret': secret}, secret


def register_complete(state, request_data):
    totp = pyotp.TOTP(state)
    if not totp.verify(request_data):
        raise ValueError
    return state


def authenticate_begin(user):
    return None, None


def authenticate_complete(state, user, request_data):
    keys = user.mfakey_set.filter(method=name)
    for key in keys:
        totp = pyotp.TOTP(key.secret)
        if (totp.verify(request_data, valid_window=settings.TOTP_VALID_WINDOW)
                and request_data != key.last_code):
            key.last_code = request_data
            key.save()
            return
    raise ValueError



from django.contrib import admin
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.views import redirect_to_login
from django.urls import reverse

from .decorators import login_not_required
from .models import MFAKey


@login_not_required
def custom_login(self, request, extra_context=None):
    next_url = (
        request.GET.get(REDIRECT_FIELD_NAME)
        or request.POST.get(REDIRECT_FIELD_NAME)
        or reverse('admin:index')
    )
    return redirect_to_login(next_url)


def patch_admin():
    setattr(admin.AdminSite, 'login', custom_login)


@admin.register(MFAKey)
class MFAKeyAdmin(admin.ModelAdmin):
    list_display = ['user', 'method', 'name']
    search_fields = ['user__username', 'name']
    list_filter = ['method']



import qrcode
import qrcode.image.svg
from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter(name='qrcode')
def get_qrcode(url):
    img = qrcode.make(url, image_factory=qrcode.image.svg.SvgImage)
    s = img.to_string().decode('utf-8')
    i = s.find('<svg')
    return mark_safe(s[i:])



try:
    from stronghold.decorators import public as stronghold_login_not_required
except ImportError:
    def stronghold_login_not_required(view_func):
        return view_func

try:
    from django.contrib.auth.decorators import login_not_required
except ImportError:
    def login_not_required(view_func):
        return view_func


def public(view_func):
    view_func.mfa_public = True
    return view_func
